/*
 * Quimeleon - a themeable quiz engine
 * Copyright © 2003, 2005, 2011  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "audio.h"

#include <iostream>
#include <cstring>



#ifdef HAVE_SOUND

# include <glibmm/timer.h>


# include <cstdio>
# include <cstdlib>
# include <unistd.h>

# include <vorbis/codec.h>

#include "gettext.h"

#endif




namespace quimeleon {




#ifdef HAVE_SOUND


const int gl_buf_size = 4096;
char gl_audio_buf[gl_buf_size]; // use memory from data segment




Glib::Thread*
init_threads()
{
	if (! Glib::thread_supported()) {
		Glib::thread_init();
	}

	return 0;
}






int Audio::_Driver = -1;
ao_device *Audio::_Device = 0;

int Audio::_Channels = 0;
int Audio::_Rate = 0;

OggVorbis_File Audio::_Oggvorbis_file;

Glib::Thread* Audio::_Thread = init_threads();

bool Audio::_Is_thread_running = false;
bool Audio::_Is_oggvorbis = false;

int Audio::_Initialized = 0;

Glib::Mutex Audio::_Mutex_common_data;


#endif





Audio::Audio()
{
#ifdef HAVE_SOUND
	if (! Glib::thread_supported()) {
		Glib::thread_init();
	}

	Glib::Mutex::Lock lock(_Mutex_common_data);
	if (! _Initialized) {
		ao_initialize();
	}
	_Initialized++;

#endif
}



Audio::~Audio()
{
#ifdef HAVE_SOUND
	Glib::Mutex::Lock lock(_Mutex_common_data);

	if (_Is_thread_running) {
		if (_Is_oggvorbis) {

			ov_clear(&_Oggvorbis_file);
			_Is_oggvorbis = false;
		}
		lock.release();

		_Thread->join();
	} else {
		lock.release();
	}


	_Initialized--;
	if (! _Initialized) {
		ao_shutdown();
	}
#endif
}




void
Audio::reset()
{
#ifdef HAVE_SOUND
	_audio_files.clear();
#endif
}



bool
Audio::play(audio_t audio)
{
#ifdef HAVE_SOUND
	const Glib::ustring filename = _audio_files[audio];
	if (! filename.empty()) {
		return play(filename);
	}
#endif

	return false;
}



bool
Audio::play(const Glib::ustring& filename)
{
#ifdef HAVE_SOUND
	try {
		Glib::Mutex::Lock lock(_Mutex_common_data);

		if (   load_file(filename)
		    && ! _Is_thread_running) {
			_Is_thread_running = true;
			_Thread = Glib::Thread::create(sigc::ptr_fun(&Audio::play_thread),
			                               true);
			_Thread->set_priority(Glib::THREAD_PRIORITY_LOW);
		}

	} catch (const Glib::Error& e) {
		std::cerr << e.what() << '\n';
	}
#endif

	return false;
}



void
Audio::mute()
{
#ifdef HAVE_SOUND
	Glib::Mutex::Lock lock(_Mutex_common_data);

	if (_Is_oggvorbis) {
		ov_clear(&_Oggvorbis_file);
		_Is_oggvorbis = false;
	}
#endif
}



void
Audio::set_sound(audio_t audio, Glib::ustring file)
{
#ifdef HAVE_SOUND
	_audio_files[audio] = file;
#endif
}



#ifdef HAVE_SOUND
// _Mutex_common_data to be locked by caller
bool
Audio::load_file(const Glib::ustring& filename) throw()
{
	FILE* fh = std::fopen(filename.c_str(), "rb");

	if (fh == 0) {
		return false;
	}


	if (_Is_oggvorbis) {
		ov_clear(&_Oggvorbis_file);
		_Is_oggvorbis = false;
	}

	if (ov_open(fh, &_Oggvorbis_file, 0, 0) < 0) {
		std::fclose(fh);
		return false;
	}
	_Is_oggvorbis = true;


	vorbis_info* vi = ov_info(&_Oggvorbis_file, -1);

	bool is_new_device = false;
	if (   ! _Device
	    || vi->channels != _Channels
	    || vi->rate != _Rate) {
		is_new_device = true;
		if (_Device) {
			ao_close(_Device);
			_Device = 0;
		}

		ao_sample_format format;
		std::memset(&format, 0, sizeof(format));
		format.bits = 16;
		format.byte_format = AO_FMT_LITTLE;
		format.channels = vi->channels;
		format.rate = vi->rate;

		// first try the last working driver, then the default driver, if
		// that fails try drivers that can handle output by multiple clients
		if (   _Driver == -1
		    || ! (_Device = ao_open_live(_Driver, &format, 0))) {
			_Driver = ao_default_driver_id();
			if (   _Driver == -1
			    || ! (_Device = ao_open_live(_Driver, &format, 0))) {
				_Driver = ao_driver_id("alsa09");
				if (   _Driver == -1
				    || ! (_Device = ao_open_live(_Driver, &format, 0))) {
					_Driver = ao_driver_id("arts");
					if (   _Driver == -1
					    || ! (_Device = ao_open_live(_Driver, &format, 0))) {
						_Driver = ao_driver_id("esd");
						if (   _Driver == -1
						   || ! (_Device = ao_open_live(_Driver, &format, 0))) {
							std::cerr << _("No suitable audio driver found")
								<< '\n';
						}
					}
				}
			}
		}

		if (_Device) {
			_Channels = format.channels;
			_Rate = format.rate;
		}
	}


	if (! _Device) {
		ov_clear(&_Oggvorbis_file);

		return false;
	}

	return is_new_device;
}



void
Audio::play_thread()
{
	int bitstream;	// out parameter of ov_read
	long len;
	for (;;) {
		{
			Glib::Mutex::Lock lock(_Mutex_common_data);

			if (! _Is_oggvorbis) {
				if (_Device) {
					ao_close(_Device);
					_Device = 0;
				}

				_Is_thread_running = false;
				break;
			}
			len = ov_read(&_Oggvorbis_file, gl_audio_buf, gl_buf_size,
			              0, 2, 1,	// little endian, 16-bit, signed
			              &bitstream);

			if (   len <= 0	// eof or error
			    || ! ao_play(_Device, gl_audio_buf, len)) {
				ov_clear(&_Oggvorbis_file);
				_Is_oggvorbis = false;

				if (_Device) {
					ao_close(_Device);
					_Device = 0;
				}

				_Is_thread_running = false;
				break;
			}
		}
		Glib::usleep(200);
	}
}
#endif





} // end namespace quimeleon
