/*
 * Quimeleon - a themeable quiz engine
 * Copyright © 2003, 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "dialogs.h"

#include <gtkmm/main.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/stock.h>

#include "about.h"
#include "configuration.h"
#include "parser.h"

#include <iostream>

#include "gettext.h"




namespace quimeleon {





Dialogs::Dialogs(Gtk::Window* parent,
	Configuration* configuration,
	Display& display, Control& control, Action& plugins)
	: _parent(parent),
	  _configuration(configuration),
	  _display(display),
	  _control(control),
	  _plugins(plugins)
{
}




void
Dialogs::file_open()
{
	static std::string last_folder;



	Gtk::FileChooserDialog file_chooser_dialog(*_parent, _8("Load file"));
	file_chooser_dialog.add_button(Gtk::Stock::CANCEL,
	                               Gtk::RESPONSE_CANCEL);
	file_chooser_dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	Gtk::FileFilter filter_quimeleon;
	filter_quimeleon.set_name(_8("Quimeleon"));
	filter_quimeleon.add_pattern("*.trivia");
	filter_quimeleon.add_pattern("*.style");
	file_chooser_dialog.add_filter(filter_quimeleon);

	Gtk::FileFilter filter_quimeleon_trivia;
	filter_quimeleon_trivia.set_name(_8("Quimeleon trivia"));
	filter_quimeleon_trivia.add_pattern("*.trivia");
	file_chooser_dialog.add_filter(filter_quimeleon_trivia);

	Gtk::FileFilter filter_quimeleon_style;
	filter_quimeleon_style.set_name(_8("Quimeleon style"));
	filter_quimeleon_style.add_pattern("*.style");
	file_chooser_dialog.add_filter(filter_quimeleon_style);

	Gtk::FileFilter filter_text;
	filter_text.set_name(_8("Text files"));
	filter_text.add_mime_type("text/plain");
	file_chooser_dialog.add_filter(filter_text);

	Gtk::FileFilter filter_any;
	filter_any.set_name(_8("All files"));
	filter_any.add_pattern("*");
	file_chooser_dialog.add_filter(filter_any);

	if (! last_folder.empty()) {
		file_chooser_dialog.set_current_folder(last_folder);
	}
	file_chooser_dialog.set_select_multiple();


	bool is_done = false;
	do {
		int re = file_chooser_dialog.run();

		switch (re) {
		case Gtk::RESPONSE_OK: {
			Configuration tmp_configuration = *_configuration;

			bool is_new_trivia = false;
			bool is_new_style = false;
			Glib::SListHandle<Glib::ustring> filenames
				= file_chooser_dialog.get_filenames();
			for (Glib::SListHandle<Glib::ustring>::const_iterator i
			     	= filenames.begin();
			     i != filenames.end(); i++) {
				std::string filename = *i;
				if (filename.size() > 6) {
					if (   filename.size() > 7
					    && filename.substr(filename.size() - 7) == ".trivia") {
						tmp_configuration.set_trivia_file(filename);
						is_new_trivia = true;
					} else if (filename.substr(filename.size() - 6) == ".style") {
						tmp_configuration.set_style_file(filename);
						is_new_style = true;
					} else {
						std::cerr << _("unknown file") << ' '
						          << filename << '\n';
					}
				}
			}

			std::string folder = file_chooser_dialog.get_current_folder();
			tmp_configuration.add_data_dir(folder);


			try {
				if (is_new_style) {
					parse_style(tmp_configuration);
				}

				if (   is_new_trivia
				    && Trivia::read_trivia_file(
				       	tmp_configuration.trivia_file().c_str(),
				       	tmp_configuration).empty()) {
					Gtk::MessageDialog message_dialog(*_parent,
						_8("Error reading trivia file"),
						false, Gtk::MESSAGE_ERROR);
					message_dialog.run();
					continue;
				}

				*_configuration = tmp_configuration;

				if (is_new_trivia) {
					apply_trivia();
				}
				if (   is_new_style) {
					apply_style();
				}

				is_done = true;

			} catch (const Exception& e) {
				Gtk::MessageDialog message_dialog(file_chooser_dialog,
					Glib::locale_to_utf8(e.what()), false,
					Gtk::MESSAGE_ERROR);
				message_dialog.run();
			}

			break;
		}
		default:
			is_done = true;
			break;
		}

		last_folder = file_chooser_dialog.get_current_folder();
	} while (! is_done);

}




void
Dialogs::preferences()
{
	Configuration tmp_configuration = *_configuration;


	Preferences preferences(*_parent, &tmp_configuration);
	preferences.show();

	do {
		int re = preferences.run();

		if (re == Gtk::RESPONSE_OK) {
			try {
				// test if style is valid
				if (preferences.is_new_style()) {
				    parse_style(tmp_configuration);
				}

				// test if trivia is valid
				if (preferences.is_new_trivia()) {
				    Trivia::read_trivia_file(
				    	tmp_configuration.trivia_file().c_str(),
				    	tmp_configuration);
				}

				*_configuration = tmp_configuration;

				if (preferences.is_new_style()) {
					apply_style();	// throws LayoutError
				}
				if (preferences.is_new_trivia()) {
					apply_trivia();
				}

				_configuration->save_style_and_trivia();
				_configuration->save_options();

				break;
			} catch (const LayoutError& e) {
				Gtk::MessageDialog message_dialog(preferences,
					Glib::locale_to_utf8(e.what()), false,
					Gtk::MESSAGE_ERROR, Gtk::BUTTONS_NONE);
				message_dialog.add_button(Gtk::Stock::QUIT, Gtk::RESPONSE_OK);
					
				message_dialog.run();

				Gtk::Main::quit();
				break;
			} catch (const Exception& e) {
				Gtk::MessageDialog(preferences,
					Glib::locale_to_utf8(e.what()), false,
					Gtk::MESSAGE_ERROR).run();
			}

		} else {

			break;
		}
	} while (1);
} // end preferences



void
Dialogs::about()
{
	About().run();
}



void
Dialogs::save_style_and_trivia()
{
	try {
		_configuration->save_style_and_trivia();
	} catch (const Exception& e) {
		Gtk::MessageDialog message_dialog(*_parent,
			Glib::locale_to_utf8(e.what()), false, Gtk::MESSAGE_ERROR);
		message_dialog.run();
	}
}



void
Dialogs::apply_style()
{
	_configuration->reset_style();
	parse_style(*_configuration);
	_plugins.reinitialize(_configuration, &_control);

	_control.on_item_top_changed();
	_display.set_item_top(_configuration->item_top());
}



void
Dialogs::apply_trivia()
{
	_control.set_trivia(
		Trivia::read_trivia_file(_configuration->trivia_file().c_str(),
		*_configuration));
	// if the number of levels changed, their size in display changed, too:
	_display.queue_resize();
}






} // end namespace quimeleon
