/*
 * Quimeleon - a themeable quiz engine
 * Copyright © 2005  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "item_pixbuf.h"

#include "gettext.h"




namespace quimeleon {





ItemPixbuf::ItemPixbuf(const Item* parent,
	const Glib::ustring& id, item_type_t type)
	: ItemArea(parent, id, type, true)
{
}




void
ItemPixbuf::set_parent_dimensions(
	int parent_x, int parent_y, int parent_w, int parent_h,
	int crop_x, int crop_y, int crop_w, int crop_h)
{
	_sub_pixbuf.clear();
	_drawable.clear();

	ItemArea::set_parent_dimensions(
		parent_x, parent_y, parent_w, parent_h,
		crop_x, crop_y, crop_w, crop_h);
}



Glib::RefPtr<Gdk::Pixbuf>&
ItemPixbuf::get_pixbuf()
{
	if (! _sub_pixbuf) {	// if get_pixbuf() invoked before compose_pixbuf()
		throw ProgramError(_("no sub-pixbuf defined"));
	}

	return _sub_pixbuf;
}



Glib::RefPtr<Gdk::Drawable>&
ItemPixbuf::drawable()
{
	try {
		Glib::RefPtr<Gdk::Pixbuf> pixbuf = get_pixbuf();
		const int width = pixbuf->get_width();
		const int height = pixbuf->get_height();
		_drawable = Gdk::Pixmap::create(
			Item::get_drawable(), width, height, -1);
		_gc = gc();

		_drawable->draw_pixbuf(_gc, pixbuf,
			0, 0, 0, 0, width, height,
			Gdk::RGB_DITHER_NORMAL, 0, 0);

	} catch (const ProgramError& e) {
		std::cerr << e.what() << '\n';
		_drawable = Glib::RefPtr<Gdk::Drawable>();
	}

	return _drawable;
}



void
ItemPixbuf::put_drawable()
{
	if (   _drawable
	    && crop_width() > 0
	    && crop_height() > 0) {
		Glib::RefPtr<Gdk::Pixbuf> pixbuf
			= Gdk::Pixbuf::create(_drawable, _drawable->get_colormap(),
				0, 0, 0, 0, crop_width(), crop_height());

		pixbuf->composite(_sub_pixbuf,
			0, 0, _sub_pixbuf->get_width(), _sub_pixbuf->get_height(),
			0, 0,
			1, 1, Gdk::INTERP_BILINEAR, 255);
	}
}



Glib::RefPtr<Gdk::GC>&
ItemPixbuf::get_gc() throw()
{
	return _gc;
}



const Gdk::Color
ItemPixbuf::color_normal()
{
	Gdk::Color color;
	color.set_rgb(_color_normal_r, _color_normal_g, _color_normal_b);
	color.rgb_find_color(_gc->get_colormap());
	return color;
}



const Gdk::Color
ItemPixbuf::color_selected()
{
	Gdk::Color color;
	color.set_rgb(_color_selected_r, _color_selected_g, _color_selected_b);
	color.rgb_find_color(_gc->get_colormap());
	return color;
}



const Gdk::Color
ItemPixbuf::color_active()
{
	Gdk::Color color;
	color.set_rgb(_color_active_r, _color_active_g, _color_active_b);
	color.rgb_find_color(_gc->get_colormap());

	return color;
}



const Gdk::Color
ItemPixbuf::color_prelight()
{
	Gdk::Color color;
	color.set_rgb(_color_prelight_r, _color_prelight_g, _color_prelight_b);
	color.rgb_find_color(_gc->get_colormap());

	return color;
}



const Gdk::Color
ItemPixbuf::color_inactive()
{
	Gdk::Color color;
	color.set_rgb(_color_inactive_r, _color_inactive_g, _color_inactive_b);
	color.rgb_find_color(_gc->get_colormap());

	return color;
}



const Gdk::Color
ItemPixbuf::color_insensitive()
{
	Gdk::Color color;
	color.set_rgb(_color_insensitive_r, _color_insensitive_g, _color_insensitive_b);
	color.rgb_find_color(_gc->get_colormap());

	return color;
}



void
ItemPixbuf::compose_pixbuf()
{

	if (   visible()
	    && crop_width() > 0
	    && crop_height() > 0
	    && _pixbuf) {
		_sub_pixbuf = Gdk::Pixbuf::create_subpixbuf(
			_pixbuf, crop_x(), crop_y(), crop_width(), crop_height())->copy();

		_signal_draw_pixbuf.emit(id());

		_sub_pixbuf->composite(_pixbuf,
			crop_x(), crop_y(), crop_width(), crop_height(), crop_x(), crop_y(),
			1, 1, Gdk::INTERP_BILINEAR, 255);

		Item::compose_pixbuf();	// not ItemRect::compose_pixbuf()
	}

}





} // end namespace quimeleon
