/*
 * Quimeleon - a themeable quiz engine
 * Copyright © 2006  Aike Reyer <aike@users.sourceforge.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include "list_view_text.h"

#include <iostream>




namespace quimeleon {





ListViewText::ListViewText()
	: _tree_model(Gtk::ListStore::create(_column_record)),
	  _is_edit_enabled(false)
{
	set_model(_tree_model);

	set_headers_visible(false);

	int n = append_column_editable("", _column_record._text);
	Gtk::TreeViewColumn* column = get_column(n - 1);

	Gtk::CellRendererText* renderer =
		dynamic_cast<Gtk::CellRendererText*>(get_column_cell_renderer(n - 1));


	renderer->property_background_gdk()
		= get_style()->get_bg(Gtk::STATE_INSENSITIVE);
	renderer->property_foreground_gdk()
		= get_style()->get_fg(Gtk::STATE_INSENSITIVE);
	column->add_attribute(renderer->property_background_set(),
		_column_record._is_marked);
	column->add_attribute(renderer->property_foreground_set(),
		_column_record._is_marked);
	column->add_attribute(renderer->property_editable(),
		_column_record._is_editable);


	Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection();
	selection->signal_changed().connect(
		sigc::mem_fun(*this, &ListViewText::on_selection_changed));
	selection->set_select_function(
		sigc::mem_fun(*this, &ListViewText::select_function));

}




void
ListViewText::append_disabled_text(const Glib::ustring& text)
{
	Gtk::TreeModel::Row row = *(_tree_model->append());

	row[_column_record._text] = text;
	row[_column_record._is_marked] = true;
	row[_column_record._is_editable] = false;

}



void
ListViewText::append_text(const Glib::ustring& text)
{
	Gtk::TreeModel::Row row = *(_tree_model->append());

	row[_column_record._text] = text;
	row[_column_record._is_marked] = false;
	row[_column_record._is_editable] = true;

}



void
ListViewText::insert_text(const Glib::ustring& text)
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();
	if (   iter
	    && ! (*iter)[_column_record._is_editable]) {
		while (   iter
		       && ! (*iter)[_column_record._is_editable]) {
			iter--;
		}
	}

	if (iter) {
		iter = _tree_model->insert_after(iter);
	} else {
		iter = _tree_model->prepend();
	}

	(*iter)[_column_record._text] = text;
	(*iter)[_column_record._is_marked] = false;
	(*iter)[_column_record._is_editable] = true;

	get_selection()->select(iter);
}



void
ListViewText::erase_selected()
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();
	if (   iter
	    && (*iter)[_column_record._is_editable]) {
		_tree_model->erase(iter);
	}
}



void
ListViewText::move_selected_up()
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();
	if (   iter
	    && iter != _tree_model->children().begin()
	    && (*iter)[_column_record._is_editable]) {
		Gtk::TreeModel::iterator iter_prev = iter;
		iter_prev--;

		_tree_model->iter_swap(iter, iter_prev);
	}
}



void
ListViewText::move_selected_down()
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();
	if (   iter
	    && (*iter)[_column_record._is_editable]) {
		Gtk::TreeModel::iterator iter_next = iter;
		iter_next++;

		if (   (*iter_next)[_column_record._is_editable]
		    && iter_next != _tree_model->children().end()) {
			_tree_model->iter_swap(iter, iter_next);
		}
	}
}



void
ListViewText::set_selected_text(const Glib::ustring& text)
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();
	if (   iter
	    && (*iter)[_column_record._is_editable]) {
		(*iter)[_column_record._text] = text;
	}
}



const Glib::ustring
ListViewText::get_selected_text()
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();
	if (iter) {
		return (*iter)[_column_record._text];
	} else {
		return "";
	}
}



void
ListViewText::on_selection_changed()
{
	Gtk::TreeModel::iterator iter = get_selection()->get_selected();

	if (   iter
	    && (*iter)[_column_record._is_editable]) {
		if (! _is_edit_enabled) {
			_signal_edit_enabled.emit(true);
			_is_edit_enabled = true;
		}
	} else if (_is_edit_enabled) {
		_signal_edit_enabled.emit(false);
		_is_edit_enabled = false;
	}

}



bool
ListViewText::select_function(const Glib::RefPtr<Gtk::TreeModel>& model,
	const Gtk::TreeModel::Path& path, bool path_currently_selected)
{
	if (path_currently_selected) {
		return true;
	}

	Gtk::TreeModel::iterator iter = model->get_iter(path);
	if (iter) {
		return (*iter)[_column_record._is_editable];
	} else {
		return false;
	}
}





} // end namespace quimeleon
