/*
 * Quimeleon - a themeable quiz engine
 * Copyright © 2005, 2006  Aike Reyer <aike@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#include <iostream>
#include <sstream>
#include <vector>

#include <libxml++/libxml++.h>

#include <gdk/gdkkeysyms.h>

#include "configuration.h"
#include "item_pixmap.h"
#include "item_rect.h"
#include "item_status.h"
#include "item_label.h"
#include "item_top.h"
#include "item_answer.h"
#include "item_action.h"
#include "item_level.h"
#include "item_question.h"
#include "item_area.h"
#include "item_pixbuf.h"
#include "parser.h"

#include "gettext.h"




namespace quimeleon {





void
verify_attribute(const Glib::ustring& expected, const Glib::ustring& value)
{
	if (expected != value) {
		throw ParserStyleError(
			std::string(_("unknown attribute")) + ' ' + value);
	}
}



void
verify_value(const Glib::ustring& expected, const Glib::ustring& value)
{
	if (expected != value) {
		throw ParserStyleError(
			std::string(_("unexpected value")) + ' ' + value);
	}
}



void
verify_tag(const Glib::ustring& expected, const Glib::ustring& value)
{
	if (expected != value) {
		throw ParserStyleError(
			std::string(_("unknown tag")) + ' ' + value);
	}
}



guint
get_key(const Glib::ustring& str)	// str.length() > 1
{
	switch (str[0]) {
	case 'A':	// Alt_L
		verify_value("Alt_L", str);
		return GDK_Alt_L;
	case 'B':	// BackSpace
		verify_value("BackSpace", str);
		return GDK_BackSpace;
	case 'C':	// Caps_Lock + Control_L + Control_R
		if (str.length() > 8) {
			switch (str[8]) {
			case 'k':	// Caps_Lock
				verify_value("Caps_Lock", str);
				return GDK_Caps_Lock;
			case 'L':	// Control_L
				verify_value("Control_L", str);
				return GDK_Control_L;
			case 'R':	// Control_R
				verify_value("Control_R", str);
				return GDK_Control_R;
			}
		}
		break;
	case 'D':	// Delete + Down
		switch (str[1]) {
		case 'e':	// Delete
			verify_value("Delete", str);
			return GDK_Delete;
		case 'o':	// Down
			verify_value("Down", str);
			return GDK_Down;
		}
		break;
	case 'E':	// End + Escape
		switch (str[1]) {
		case 'n':	// End
			verify_value("End", str);
			return GDK_End;
		case 's':	// Escape
			verify_value("Escape", str);
			throw ParserStyleError(
				_("redefinition of quit-key (Escape) not allowed"));
		}
		break;
	case 'F': {
		std::istringstream istr(str.substr(1));
		int n;
		istr >> n;

		if (n > 0) {
			return GDK_F1 + n - 1;
		}

		break;
	}
	case 'H':	// Home
		verify_value("Home", str);
		return GDK_Home;
	case 'I':	// Insert + ISO_Level3_Shift
		switch (str[1]) {
		case 'n':	// Insert
			verify_value("Insert", str);
			return GDK_Insert;
		case 'S':	// ISO_Level3_Shift
			verify_value("ISO_Level3_Shift", str);
			return GDK_ISO_Level3_Shift;
		}
		break;
	case 'K':
		if (str.length() > 3) {
			switch (str[3]) {
			case '0':	// KP_0
				verify_value("KP_0", str);
				return GDK_KP_0;
			case '1':	// KP_1
				verify_value("KP_1", str);
				return GDK_KP_1;
			case '2':	// KP_2
				verify_value("KP_2", str);
				return GDK_KP_2;
			case '3':	// KP_3
				verify_value("KP_3", str);
				return GDK_KP_3;
			case '4':	// KP_4
				verify_value("KP_4", str);
				return GDK_KP_4;
			case '5':	// KP_5
				verify_value("KP_5", str);
				return GDK_KP_5;
			case '6':	// KP_6
				verify_value("KP_6", str);
				return GDK_KP_6;
			case '7':	// KP_7
				verify_value("KP_7", str);
				return GDK_KP_7;
			case '8':	// KP_8
				verify_value("KP_8", str);
				return GDK_KP_8;
			case '9':	// KP_9
				verify_value("KP_9", str);
				return GDK_KP_9;
			case 'A':	// KP_Add
				verify_value("KP_Add", str);
				return GDK_KP_Add;
			case 'B':	// KP_Begin
				verify_value("KP_Begin", str);
				return GDK_KP_Begin;
			case 'D':	// KP_Delete + KP_Divide + KP_Down
				if (str.length() > 4) {
					switch (str[4]) {
					case 'e':	// KP_Delete
						verify_value("KP_Delete", str);
						return GDK_KP_Delete;
					case 'i':	// KP_Divide
						verify_value("KP_Divide", str);
						return GDK_KP_Divide;
					case 'o':	// KP_Down
						verify_value("KP_Down", str);
						return GDK_KP_Down;
					}
				}
				break;
			case 'E':	// KP_End + KP_Enter
				if (str.length() > 5) {
					switch (str[5]) {
					case 'd':	// KP_End
						verify_value("KP_End", str);
						return GDK_KP_End;
					case 't':	// KP_Enter
						verify_value("KP_Enter", str);
						return GDK_KP_Enter;
					}
				}
				break;
			case 'H':	// KP_Home
				verify_value("KP_Home", str);
				return GDK_KP_Home;
			case 'I':	// KP_Insert
				verify_value("KP_Insert", str);
				return GDK_KP_Insert;
			case 'L':	// KP_Left:
				verify_value("KP_Left", str);
				return GDK_KP_Left;
			case 'M':	// KP_Multiply
				verify_value("KP_Multiply", str);
				return GDK_KP_Multiply;
			case 'N':	// KP_Next
				verify_value("KP_Next", str);
				return GDK_KP_Next;
			case 'P':	// KP_Page_Down + KP_Page_Up
				if (str.length() > 8) {
					switch (str[8]) {
					case 'D':	// KP_Page_Down
						verify_value("KP_Page_Down", str);
						return GDK_KP_Page_Down;
					case 'U':	// KP_Page_Up
						verify_value("KP_Page_Up", str);
						return GDK_KP_Page_Up;
					}
				}
				break;
			case 'R':	// KP_Right
				verify_value("KP_Right", str);
				return GDK_KP_Right;
			case 'S':	// KP_Separator
				if (str.length() > 4) {
					switch (str[4]) {
					case 'e':	// KP_Separator
						verify_value("KP_Separator", str);
						return GDK_KP_Separator;
					case 'u':	// KP_Subtract
						verify_value("KP_Subtract", str);
						return GDK_KP_Subtract;
					}
				}
				break;
			case 'U':	// KP_Up
				verify_value("KP_Up", str);
				return GDK_KP_Up;
			}
		}
		break;
	case 'L':	// Left
		verify_value("Left", str);
		return GDK_Left;
	case 'M':	// Menu
		verify_value("Menu", str);
		return GDK_Menu;
	case 'N':	// Next + Num_Lock
		switch (str[1]) {
		case 'e':	// Next
			verify_value("Next", str);
			return GDK_Next;
		case 'u':	// Num_Lock
			verify_value("Num_Lock", str);
			return GDK_Num_Lock;
		}
		break;
	case 'P':	// Page_Up + Page_Down
		if (str.length() > 5) {
			switch (str[5]) {
			case 'D':	// Page_Down
				verify_value("Page_Down", str);
				return GDK_Page_Down;
			case 'U':	// Page_Up
				verify_value("Page_Up", str);
				return GDK_Page_Up;
			}
		} else if (str.length() > 3) {
			switch (str[3]) {
			case 's':	// Pause
				verify_value("Pause", str);
				return GDK_Pause;
			case 'n':	// Print
				verify_value("Print", str);
				return GDK_Print;
			case 'o':	// Prior
				verify_value("Prior", str);
				return GDK_Prior;
			}
		}
		break;
	case 'R':
		switch (str[1]) {
		case 'e':	// Return
			verify_value("Return", str);
			return GDK_Return;
		case 'i':	// Right
			verify_value("Right", str);
			return GDK_Right;
		}
	case 'S':	// Scroll_Lock + Shift_L + Shift_R + Super_L + Super_R
		if (str.length() > 6) {
			switch (str[1]) {
			case 'c':	// Scroll_Lock
				verify_value("Scroll_Lock", str);
				return GDK_Scroll_Lock;
			case 'h':	// Shift_L + Shift_R
				switch (str[6]) {
				case 'L':	// Shift_L
					verify_value("Shift_L", str);
					return GDK_Shift_L;
				case 'R':	// Shift_R
					verify_value("Shift_R", str);
					return GDK_Shift_R;
				}
				break;
			case 'u':	// Super_L + Super_R
				switch (str[6]) {
				case 'L':	// Super_L
					verify_value("Super_L", str);
					return GDK_Super_L;
				case 'R':	// Super_R
					verify_value("Super_R", str);
					return GDK_Super_R;
				}
				break;
			}
		}
		break;
	case 'T':	// Tab
		verify_value("Tab", str);
		return GDK_Tab;
	case 'U':	// Up
		verify_value("Up", str);
		return GDK_Up;
	}

	throw ParserStyleError(std::string(_("no such key")) + ": " + str);
}




void
parse_item_attributes_id(const xmlpp::Element::AttributeList& attribute_list,
	Glib::ustring& id)
{
	for (xmlpp::Element::AttributeList::const_iterator i
	     	= attribute_list.begin();
	     i != attribute_list.end(); i++) {

		Glib::ustring attribute_value = (*i)->get_value();
		Glib::ustring attribute_name = (*i)->get_name();
		if (attribute_value.empty()) {
			throw ParserStyleError(
				attribute_name + _(": no value"));
		}

		switch (attribute_name[0]) {
		case 'i':	// id
			verify_attribute("id", attribute_name);
			id = attribute_value;
			break;
		default:
			throw ParserStyleError(
				  std::string(_("unknown attribute"))
				+ ": " + attribute_name);
		}
	}
}



void
parse_item_attributes_id_number(
	const xmlpp::Element::AttributeList& attribute_list,
	Glib::ustring& id,
	unsigned int& number)
{
	for (xmlpp::Element::AttributeList::const_iterator i
	     	= attribute_list.begin();
	     i != attribute_list.end(); i++) {

		Glib::ustring attribute_value = (*i)->get_value();
		Glib::ustring attribute_name = (*i)->get_name();
		if (attribute_value.empty()) {
			throw ParserStyleError(
				attribute_name + _(": no value"));
		}

		switch (attribute_name[0]) {
		case 'i':	// id
			verify_attribute("id", attribute_name);
			id = attribute_value;
			break;
		case 'n': {	// number
			verify_attribute("number", attribute_name);
			std::istringstream istr(attribute_value);
			istr >> number;
			break;
		}
		default:
			throw ParserStyleError(
				  std::string(_("unknown attribute"))
				+ ": " + attribute_name);
		}
	}
}




void
recurse_node_display(const xmlpp::Node* node, Item* item, group_t group,
	const Configuration* configuration)
{
	item_type_t item_type = T_unknown;

	if (const xmlpp::ContentNode* content_node
	    	= dynamic_cast<const xmlpp::ContentNode*>(node)) {
		if (   ! dynamic_cast<const xmlpp::CommentNode*>(node)
		    && ! content_node->is_white_space()) {
			item->set_content(content_node->get_content());
		}

	} else if (const xmlpp::Element* element
	           	= dynamic_cast<const xmlpp::Element*>(node)) {
		Glib::ustring name = element->get_name();
		bool is_alternatives = false;
		bool is_sequence = false;

		switch (name[0]) {
		case 'd': {	// draw
			verify_tag("draw", name);
			item_type = T_draw;

			Glib::ustring id;
			parse_item_attributes_id(element->get_attributes(), id);

			Item* child = new ItemPixbuf(item, id, item_type);

			if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
				item_top->add_child(child, group);
			} else {
				item->add_child(child);
			}

			item = child;

			break;
		}
		case 'v':	// varpixmap + varlabel
			if (name.length() > 3) {
				switch (name[3]) {
				case 'p':
					verify_tag("varpixmap", name);
					item_type = T_varpixmap;
					goto parse_pixmap;
				case 'l':
					verify_tag("varlabel", name);
					item_type = T_varlabel;
					goto parse_label;
				default:
					throw ParserStyleError(
						  std::string(_("unknown tag"))
						+ ": " + name);
				}
			} else {
				throw ParserStyleError(
					  std::string(_("unknown tag"))
					+ ": " + name);
			}

			break;
		case 'l': {	// label + level
			if (name.length() > 1) {
				switch (name[1]) {
				case 'a': {
					verify_tag("label", name);
					item_type = T_label;
				parse_label:

					Glib::ustring id;
					parse_item_attributes_id(element->get_attributes(), id);
					Item* child = new ItemLabel(item, id, item_type);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(child, group);
					} else {
						item->add_child(child);
					}

					item = child;

					break;
				}
				case 'e': {	// level
					verify_tag("level", name);
					item_type = T_rect;

					Glib::ustring id;
					unsigned int number = 0;
					parse_item_attributes_id_number(element->get_attributes(),
						id, number);

					ItemLevel* child
						= new ItemLevel(item, id, item_type, number);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(static_cast<Item*>(child), group);
					} else {
						item->add_child(static_cast<Item*>(child));
					}

					item = child;

					break;
				}
				}
			} else {
				throw ParserStyleError(
					  std::string(_("unknown tag"))
					+ ": " + name);
			}

			break;
		}
		case 'p':	// pixmap + pause + pos + prelight
			if (name.length() > 1) {
				switch (name[1]) {
				case 'i': {	// pixmap
					verify_tag("pixmap", name);
					item_type = T_pixmap;
				parse_pixmap:

					Glib::ustring id;
					bool keep_aspect = false;
					const xmlpp::Element::AttributeList& attribute_list
						= element->get_attributes();
					for (xmlpp::Element::AttributeList::const_iterator i
					     	= attribute_list.begin();
					     i != attribute_list.end(); i++) {

						Glib::ustring attribute_value = (*i)->get_value();
						Glib::ustring attribute_name = (*i)->get_name();
						if (attribute_value.empty()) {
							throw ParserStyleError(
								attribute_name + _(": no value"));
						}

						switch (attribute_name[0]) {
						case 'i':	// id
							verify_attribute("id", attribute_name);
							id = attribute_value;
							break;
						case 'k':	// keepaspect
							verify_attribute("keepaspect", attribute_name);
							if (attribute_value[0] == 't') {
								verify_value("true", attribute_value);
								keep_aspect = true;
							} else {
								verify_value("false", attribute_value);
								keep_aspect = false;
							}
							break;
						default:
							throw ParserStyleError(
								  std::string(_("unknown attribute"))
								+ ": " + attribute_name);
						}
					}

					Item* child
						= new ItemPixmap(item, id, item_type, keep_aspect,
						                 configuration->do_buffer());

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(child, group);
					} else {
						item->add_child(child);
					}

					item = child;

					break;
				}
				case 'r':	// prelight
					verify_tag("prelight", name);
					goto read_state;
				case 'a': {	// pause
					verify_tag("pause", name);
					unsigned int seconds = 0;

					const xmlpp::Element::AttributeList& attribute_list
						= element->get_attributes();

					for (xmlpp::Element::AttributeList::const_iterator i
					     	= attribute_list.begin();
					     i != attribute_list.end(); i++) {

						Glib::ustring attribute_name = (*i)->get_name();
						Glib::ustring attribute_value = (*i)->get_value();
						if (attribute_value.empty()) {
							throw ParserStyleError(
								attribute_name + _(": no value"));
						}

						verify_attribute("seconds", attribute_name);
						std::istringstream istr(attribute_value);
						istr >> seconds;
					}

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_pause(seconds, group);
					} else {
						throw ParserStyleError(
							_("<pause> only allowed within <sequence>"));
					}
					break;
				}
				default:	// pos
					verify_tag("pos", name);
					const xmlpp::Element::AttributeList& attribute_list
						= element->get_attributes();

					for (xmlpp::Element::AttributeList::const_iterator i
					     	= attribute_list.begin();
					     i != attribute_list.end(); i++) {

						Glib::ustring attribute_name = (*i)->get_name();
						Glib::ustring attribute_value = (*i)->get_value();
						if (attribute_value.empty()) {
							throw ParserStyleError(
								attribute_name + _(": no value"));
						}

						switch (attribute_name[0]) {
						case 'a': {	// anchor
							verify_attribute("anchor", attribute_name);
							anchor_t anchor;
							switch (attribute_value[0]) {
							case 'n':	// n + ne + nw
								if (attribute_value.length() > 1) {
									if (attribute_value[1] == 'e') {
										verify_value("ne", attribute_value);
										anchor = NE;
									} else {
										verify_value("nw", attribute_value);
										anchor = NW;
									}
								} else {
									verify_value("n", attribute_value);
									anchor = N;
								}
								break;
							case 'e':	// e
								verify_value("e", attribute_value);
								anchor = E;
								break;
							case 's':	// s + se + sw
								if (attribute_value.length() > 1) {
									if (attribute_value[1] == 'e') {
										verify_value("se", attribute_value);
										anchor = SE;
									} else {
										verify_value("sw", attribute_value);
										anchor = SW;
									}
								} else {
									verify_value("s", attribute_value);
									anchor = S;
								}
								break;
							case 'w':	// w
								verify_value("w", attribute_value);
								anchor = W;
								break;
							case 'c':	// c
								verify_value("c", attribute_value);
								anchor = C;
								break;
							default:
								verify_value("nw", attribute_value);
								anchor = NW;
							}

							item->set_anchor(anchor);
							break;
						}
						case 'x':
						case 'y':
						case 'w':
						case 'h': {
							bool is_percent = false;
							if (attribute_value[attribute_value.length() - 1]
							    	== '%') {
								is_percent = true;
							}
							std::istringstream istr(attribute_value);
							int n;
							istr >> n;

							switch (attribute_name[0]) {
							case 'x':
								verify_attribute("x", attribute_name);
								item->set_x(n, is_percent);
								break;
							case 'y':
								verify_attribute("y", attribute_name);
								item->set_y(n, is_percent);
								break;
							case 'w':
								verify_attribute("width", attribute_name);
								item->set_width(n, is_percent);
								break;
							case 'h':
								verify_attribute("height", attribute_name);
								item->set_height(n, is_percent);
								break;
							}

							break;
						}
						default:
							throw ParserStyleError(
								  std::string(_("unknown attribute"))
								+ ": " + attribute_name);
						}
					} // end for
				} // end switch (name[1])

			} // end if (name.length() > 1) {
			break;
		case 'a':	// area + answer + action + active + alternatives
			if (name.length() > 1) {
				switch (name[1]) {
				case 'r': {	// area
					verify_tag("area", name);
					item_type = T_area;

					Glib::ustring id;
					parse_item_attributes_id(element->get_attributes(), id);

					Item* child = new ItemArea(item, id, item_type);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(child, group);
					} else {
						item->add_child(child);
					}

					item = child;

					break;
				}
				case 'n': {	// answer
					verify_tag("answer", name);
					item_type = T_rect;

					Glib::ustring id;
					unsigned int number = 0;
					parse_item_attributes_id_number(element->get_attributes(),
						id, number);


					ItemAnswer* child
						= new ItemAnswer(item, id, item_type, number);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(static_cast<Item*>(child), group);
					} else {
						item->add_child(static_cast<Item*>(child));
					}

					item = child;

					break;
				}
				case 'c': 	// action + active
					if (name.length() > 4) {
						switch (name[4]) {
						case 'o': {	// action
							verify_tag("action", name);
							item_type = T_rect;

							Glib::ustring id;
							unsigned int number = 0;
							Glib::ustring plugin;
							Glib::ustring plugin_arg;

							const xmlpp::Element::AttributeList& attribute_list
								= element->get_attributes();
							for (xmlpp::Element::AttributeList::const_iterator i
							     	= attribute_list.begin();
							     i != attribute_list.end(); i++) {

								Glib::ustring attribute_name = (*i)->get_name();
								Glib::ustring attribute_value = (*i)->get_value();
								if (attribute_value.empty()) {
									throw ParserStyleError(
										attribute_name + _(": no value"));
								}

								switch (attribute_name[0]) {
								case 'i':	// id
									verify_attribute("id", attribute_name);
									id = attribute_value;
									break;
								case 'n': {	// number
									verify_attribute("number", attribute_name);
									std::istringstream istr(attribute_value);
									istr >> number;
									break;
								}
								case 'p':	// plugin + pluginarg
									if (attribute_name.length() > 6) {
										verify_attribute("pluginarg", attribute_name);
										plugin_arg = attribute_value;
									} else {
										verify_attribute("plugin", attribute_name);
										plugin = attribute_value;
									}
									break;
								default:
									throw ParserStyleError(
										  std::string(_("unknown attribute"))
										+ ": " + attribute_name);
								}
							}

							ItemAction* child
								= new ItemAction(item, id, item_type, number,
								                 plugin, plugin_arg);

							if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
								item_top->add_child(static_cast<Item*>(child), group);
							} else {
								item->add_child(static_cast<Item*>(child));
							}

							item = child;

							break;
						}
						case 'v':	// active
							verify_tag("active", name);
							goto read_state;
						}
					} else {
						throw ParserStyleError(
							  std::string(_("unknown tag"))
							+ ": " + name);
					}

					break;
				case 'l':	// alternatives
					verify_tag("alternatives", name);
					is_alternatives = true;
					if (group == group_trivia) {
						item->parse_alternatives(true);
					} else {
						throw ParserStyleError(
							_("<alternatives> only allowed in group trivia"));
					}
					break;
				default:
					throw ParserStyleError(
						  std::string(_("unknown tag"))
						+ ": " + name);
				}
			} else {
				throw ParserStyleError(
					  std::string(_("unknown tag"))
					+ ": " + name);
			}
			break;
		case 's':	// selected + sound + sequence + status
			if (name.length() > 2) {
				switch (name[2]) {
				case 'u': {	// sound
					verify_tag("sound", name);
					const xmlpp::Element::AttributeList& attribute_list
						= element->get_attributes();

					for (xmlpp::Element::AttributeList::const_iterator i
					     	= attribute_list.begin();
					     i != attribute_list.end(); i++) {

						Glib::ustring attribute_name = (*i)->get_name();
						Glib::ustring attribute_value = (*i)->get_value();
						if (attribute_value.empty()) {
							throw ParserStyleError(
								attribute_name + _(": no value"));
						}

						switch (attribute_name[0]) {
						case 'f':	// file
							verify_attribute("file", attribute_name);
#ifndef HAVE_SOUND
							try {
#endif
							item->set_soundfile(
								configuration->find_file(attribute_value));
#ifndef HAVE_SOUND
							} catch (const FileNotFound& e) {
								std::cerr << e.what()
									<< ' ' << _("(ignored)") << '\n';
							}
#endif
							break;
						case 'c':	// continue
							verify_attribute("continue", attribute_name);
							if (attribute_value[0] == 't') {	// true
								verify_value("true", attribute_value);
								item->set_silence(false);
							} else {	// false
								verify_value("false", attribute_value);
								item->set_silence(true);
							}
							break;
						default:
							throw ParserStyleError(
								  std::string(_("unknown attribute"))
								+ ": " + attribute_name);
						}
					}
					break;
				}
				case 'q':	// sequence
					verify_tag("sequence", name);
					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						is_sequence = true;
						item_top->parse_sequence(true, group);
					} else if (! dynamic_cast<ItemStatus*>(item)) {
						throw ParserStyleError(
							_("<sequence> only allowed in top level of groups"));
					}
					break;
				case 'a': {	// status
					verify_tag("status", name);
					item_type = T_status;

					const xmlpp::Element::AttributeList& attribute_list
						= element->get_attributes();

					Glib::ustring id;
					ItemStatus::reference_t reference = ItemStatus::R_none;
					unsigned int number = 0;

					for (xmlpp::Element::AttributeList::const_iterator i
					     	= attribute_list.begin();
					     i != attribute_list.end(); i++) {

						Glib::ustring attribute_name = (*i)->get_name();
						Glib::ustring attribute_value = (*i)->get_value();
						if (attribute_value.empty()) {
							throw ParserStyleError(
								attribute_name + _(": no value"));
						}

						switch (attribute_name[0]) {
						case 'r':	// reference
							verify_attribute("reference", attribute_name);

							switch (attribute_value[0]) {
							case 'l':	// level
								verify_value("level", attribute_value);

								reference = ItemStatus::R_level;
								break;
							case 'q':	// question
								verify_value("question", attribute_value);

								reference = ItemStatus::R_question;
								break;
							case 'n':	// next_level + next_question
							         	// + next_level_and_question
								if (attribute_value.length() > 5) {
									switch (attribute_value[5]) {
									case 'l':	// next_level
										if (attribute_value.length() > 10) {
											verify_value(
												"next_level_and_question",
												attribute_value);

											reference = ItemStatus::R_next_level_and_question;
										} else {
											verify_value("next_level",
												attribute_value);

											reference
												= ItemStatus::R_next_level;
										}
										break;
									case 'q':	// next_question
										verify_value("next_question",
											attribute_value);

										reference = ItemStatus::R_next_question;
										break;
									default:
										throw ParserStyleError(
											std::string(_("unexpected value"))
											+ ": " + attribute_value);
									}
								}
								break;
							case 'm':	// manual
								verify_value("manual", attribute_value);

								reference = ItemStatus::R_manual;
								break;
							}
							break;
						case 'i':	// id
							verify_attribute("id", attribute_name);
							id = attribute_value;
							break;
						case 'n': {	// number
							verify_attribute("number", attribute_name);
							std::istringstream istr(attribute_value);
							istr >> number;
							number--;
							break;
						}
						default:
							throw ParserStyleError(
								  std::string(_("unknown attribute"))
								+ ": " + attribute_name);
						}
					}

					ItemStatus* child
						= new ItemStatus(item, id, item_type, number,
						                 reference);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(static_cast<Item*>(child), group);
					} else {
						item->add_child(static_cast<Item*>(child));
					}

					item = child;

					break;
				}
				case 'l':	// selected
					verify_tag("selected", name);
					goto read_state;
				default:
					throw ParserStyleError(
						  std::string(_("unknown tag"))
						+ ": " + name);
				}
			} else {
				throw ParserStyleError(
					  std::string(_("unknown tag"))
					+ ": " + name);
			}
			break;
		case 'n': {	// normal
			verify_tag("normal", name);
		read_state:
			if (ItemRect* item_rect = dynamic_cast<ItemRect*>(item)) {
				item_rect->set_transparent(false);
			}

			const xmlpp::Element::AttributeList& attribute_list
				= element->get_attributes();

			for (xmlpp::Element::AttributeList::const_iterator i
			     	= attribute_list.begin();
			     i != attribute_list.end(); i++) {

				Glib::ustring attribute_name = (*i)->get_name();
				Glib::ustring attribute_value = (*i)->get_value();
				if (attribute_value.empty()) {
					throw ParserStyleError(attribute_name + _(": no value"));
				}

				switch (attribute_name[0]) {
				case 'a': {	// alpha
					verify_attribute("alpha", attribute_name);
					std::istringstream istr(attribute_value);
					int alpha;
					istr >> alpha;

					switch (name[0]) {
					case 'n':	// normal
						item->set_alpha_normal(alpha);
						break;
					case 's':	// selected
						item->set_alpha_selected(alpha);
						break;
					case 'a':	// active
						item->set_alpha_active(alpha);
						break;
					case 'p':	// prelight
						item->set_alpha_prelight(alpha);
						break;
					case 'i':	// inactive + insensitive
						switch (name[2]) {
						case 'a':	// inactive
							item->set_alpha_inactive(alpha);
							break;
						case 's':	// insensitive
							item->set_alpha_insensitive(alpha);
							break;
						}
						break;
					}
					break;
				}
				case 'c': {	// color
					verify_attribute("color", attribute_name);
					int r = 0;
					int g = 0;
					int b = 0;

					for (Glib::ustring::size_type pos = 1;
					     pos < attribute_value.length(); pos += 2) {
						std::istringstream istr(attribute_value.substr(pos, 2));
						int i;

						istr >> std::hex >> i;
						switch (pos) {
						case 1:
							r = i;
							break;
						case 3:
							g = i;
							break;
						case 5:
							b = i;
							break;
						default:
							throw ParserStyleError(
								  Glib::ustring(attribute_value) + ' '
								+ _("is not a color in hex format (too long)"));
						}
					}

					switch (name[0]) {
					case 'n':	// normal
						item->set_color_normal(r * 0xff, g * 0xff, b * 0xff);
						break;
					case 's':	// selected
						item->set_color_selected(r * 0xff, g * 0xff, b * 0xff);
						break;
					case 'a':	// active
						item->set_color_active(r * 0xff, g * 0xff, b * 0xff);
						break;
					case 'p':	// prelight
						item->set_color_prelight(r * 0xff, g * 0xff, b * 0xff);
						break;
					case 'i':	// inactive + insensitive
						switch (name[2]) {
						case 'a':
							item->set_color_inactive(
								r * 0xff, g * 0xff, b * 0xff);
							break;
						case 's':
							item->set_color_insensitive(
								r * 0xff, g * 0xff, b * 0xff);
							break;
						}
						break;
					}

					break;
				}
				case 'f':
					if (attribute_name.length() > 1) {
						switch (attribute_name[1]) {
						case 'o':	// font
							verify_attribute("font", attribute_name);
							if (ItemLabel* item_label
							    	= dynamic_cast<ItemLabel*>(item)) {
								switch (name[0]) {
								case 'n':	// normal
									item_label->set_font_normal(
										attribute_value);
									break;
								case 's':	// selected
									item_label->set_font_selected(
										attribute_value);
									break;
								case 'a':	// active
									item_label->set_font_active(
										attribute_value);
									break;
								case 'p':	// prelight
									item_label->set_font_prelight(
										attribute_value);
									break;
								case 'i':	// inactive + insensitive
									switch (name[2]) {
									case 'a':
										item_label->set_font_inactive(
											attribute_value);
										break;
									case 's':
										item_label->set_font_insensitive(
											attribute_value);
										break;
									}
									break;
								}
							}
							break;
						default:	// file
							verify_attribute("file", attribute_name);
							if (ItemPixmap* item_pixmap
							    	= dynamic_cast<ItemPixmap*>(item)) {
								Glib::ustring filename
									= configuration->find_file(attribute_value);
								switch (name[0]) {
								case 'n':	// normal
									item_pixmap->set_normal(filename);
									break;
								case 's':	// selected
									item_pixmap->set_selected(filename);
									break;
								case 'a':	// active
									item_pixmap->set_active(filename);
									break;
								case 'p':	// prelight
									item_pixmap->set_prelight(filename);
									break;
								case 'i':	// inactive + insensitive
									switch (name[2]) {
									case 'a':
										item_pixmap->set_inactive(filename);
										break;
									case 's':
										item_pixmap->set_insensitive(filename);
										break;
									}
									break;
								}
							}
							break;
						} // end switch
					} else {
						throw ParserStyleError(
							  std::string(_("unknown attribute"))
							+ ": " + name);
					}
					break;
				}
			}

			break;
		}

		case 'i': // inactive + insensitive
			if (name.length() > 2) {
				if (name[2] == 'a') {
					verify_tag("inactive", name);
				} else {
					verify_tag("insensitive", name);
				}
				goto read_state;	// no need to check for name.length() > 2
				                	// at read_state
			} else {
				throw ParserStyleError(
					  std::string(_("unknown tag"))
					+ ": " + name);
			} // end if (name.length() > 2)

			break; // end case 'i'
		case 't':	// text
			if (ItemLabel* item_label = dynamic_cast<ItemLabel*>(item)) {
				const xmlpp::Element::AttributeList& attribute_list
					= element->get_attributes();
				for (xmlpp::Element::AttributeList::const_iterator i
				     	= attribute_list.begin();
				     i != attribute_list.end(); i++) {

					Glib::ustring attribute_name = (*i)->get_name();
					Glib::ustring attribute_value = (*i)->get_value();
					if (attribute_value.empty()) {
						throw ParserStyleError(
							attribute_name + _(": no value"));
					}

					switch (attribute_name[0]) {
					case 'a':	// alignment
						verify_attribute("alignment", attribute_name);
						switch (attribute_value[0]) {
						case 'r':
							verify_value("right", attribute_value);
							item_label->set_alignment(align_right);
							break;
						case 'c':
							verify_value("center", attribute_value);
							item_label->set_alignment(align_center);
							break;
						case 'j':
							verify_value("justify", attribute_value);
							item_label->set_alignment(align_justify);
							break;
						case 'l':
							verify_attribute("left", attribute_name);
							item_label->set_alignment(align_left);
							break;
						default:
							throw ParserStyleError(
								  std::string(_("Unknown alignment: "))
								+ attribute_value);
						}
						break;
					case 'p':	// parse
						verify_attribute("parse", attribute_name);
						if (attribute_value[0] == 't') {	// true
							verify_value("true", attribute_value);
							item_label->set_parse(true);
						} else {	// false
							verify_value("false", attribute_value);
							item_label->set_parse(false);
						}
						break;
					default:
						throw ParserStyleError(
							  std::string(_("unknown attribute"))
							+ ": " + attribute_name);
					}
				}
			}
			break;
		case 'f':	// frame + fixed + filler
			if (name.length() > 2) {
				switch (name[2]) {
				case 'x':	// fixed
					verify_tag("fixed", name);
					if (ItemRect* item_rect = dynamic_cast<ItemRect*>(item)) {
						item_rect->set_fixed(true);
					}

					break;
				case 'a': {	// frame
					verify_tag("frame", name);
					item_type = T_rect;

					Glib::ustring id;
					parse_item_attributes_id(element->get_attributes(), id);

					Item* child = new ItemRect(item, id, item_type);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(child, group);
					} else {
						item->add_child(child);
					}

					item = child;

					break;
				}
				case 'l':	// filler
					verify_tag("filler", name);
					if (ItemStatus* item_status
					    	= dynamic_cast<ItemStatus*>(item)) {
						unsigned int count = 1;

						const xmlpp::Element::AttributeList& attribute_list
							= element->get_attributes();
						for (xmlpp::Element::AttributeList::const_iterator i
						     	= attribute_list.begin();
						     i != attribute_list.end(); i++) {
		
							Glib::ustring attribute_name = (*i)->get_name();
							Glib::ustring attribute_value = (*i)->get_value();
							if (attribute_value.empty()) {
								throw ParserStyleError(
									attribute_name + _(": no value"));
							}
		
							switch (attribute_name[0]) {
							case 'c': {	// count
								verify_attribute("count", attribute_name);
								std::istringstream istr(attribute_value);
								istr >> count;

								break;
							}
							default:
								throw ParserStyleError(
									  std::string(_("unknown attribute"))
									+ ": " + attribute_name);
							}
						}

						for (unsigned int i = 0; i < count; i++) {
							item_status->add_filler(0);
						}
					} else {
						throw ParserStyleError(
							_("<filler> only allowed as child of <status>"));
					}
					break;
				default:
					throw ParserStyleError(
						  std::string(_("unknown tag"))
						+ ": " + name);
				}
			} else {
				throw ParserStyleError(
					  std::string(_("unknown tag"))
					+ ": " + name);
			}

			break;
		case 'q': {	// quimeleon + question

			if (name.length() > 2) {
				switch (name[2]) {
				case 'i':	// quimeleon
					verify_tag("quimeleon", name);
					break;
				case 'e': {	// question
					verify_tag("question", name);
					item_type = T_rect;

					Glib::ustring id;
					parse_item_attributes_id(element->get_attributes(), id);

					Item* child = new ItemQuestion(item, id, item_type);

					if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
						item_top->add_child(child, group);
					} else {
						item->add_child(child);
					}

					item = child;

					break;
				}
				default:
					throw ParserStyleError(
						  std::string(_("unknown tag"))
						+ ": " + name);
				}
			}
			break;
		}
		case 'b':	// box
			verify_tag("box", name);
			if (ItemRect* item_rect = dynamic_cast<ItemRect*>(item)) {
				item_rect->set_fixed(false);
				const xmlpp::Element::AttributeList& attribute_list
					= element->get_attributes();
				for (xmlpp::Element::AttributeList::const_iterator i
				     	= attribute_list.begin();
				     i != attribute_list.end(); i++) {

					Glib::ustring attribute_name = (*i)->get_name();
					Glib::ustring attribute_value = (*i)->get_value();
					if (attribute_value.empty()) {
						throw ParserStyleError(
							attribute_name + _(": no value"));
					}

					switch (attribute_name[0]) {
					case 'o':	// orientation
						verify_attribute("orientation", attribute_name);
						if (attribute_value[0] == 'h') {	// horizontal
							verify_value("horizontal", attribute_value);
							item_rect->set_horizontal(true);
						} else {	// vertical
							verify_value("vertical", attribute_value);
							item_rect->set_horizontal(false);
						}
						break;
					case 'p':	// padding
						verify_attribute("padding", attribute_name);
						break;
					case 's': {	// spacing
						verify_attribute("spacing", attribute_name);

						if (attribute_value[attribute_value.length() - 1]
						    	!= '%') {
							throw ParserStyleError(
								_("spacing must be in percent"));
						}
						std::istringstream istr(attribute_value);
						int n;
						istr >> n;

						item_rect->set_spacing(n);

						break;
					}
					case 'm':	// min
						verify_attribute("min", attribute_name);
						break;
					default:
						throw ParserStyleError(
							  std::string(_("unknown attribute"))
							+ ": " + attribute_name);
					}
				}
			}
			break;
		default:
			throw ParserStyleError(
				  std::string(_("unknown tag"))
				+ ": " + name);
		} // end switch (name[0])


		xmlpp::Node::NodeList node_list = element->get_children();
		for (xmlpp::Node::NodeList::iterator iter = node_list.begin();
			 iter != node_list.end(); iter++) {
			recurse_node_display(*iter, item, group, configuration);
		}

		if (is_alternatives) {
			item->parse_alternatives(false);
		} else if (is_sequence) {
			if (ItemTop* item_top = dynamic_cast<ItemTop*>(item)) {
				item_top->parse_sequence(false, group);
			}
		}

		if (item_type != T_unknown) {
			item->complete_attributes();
		}
	} // end if content_node else if element

}



void
recurse_node_audio(const xmlpp::Node* node, Audio* audio,
	const Configuration* configuration)
{
	if (const xmlpp::Element* element
	    	= dynamic_cast<const xmlpp::Element*>(node)) {
		Glib::ustring name = element->get_name();

		switch (name[0]) {
		case 's': {	// sound
			verify_tag("sound", name);
			Audio::audio_t id = Audio::unknown;
			Glib::ustring file;

			const xmlpp::Element::AttributeList& attribute_list
				= element->get_attributes();

			for (xmlpp::Element::AttributeList::const_iterator i
			     	= attribute_list.begin();
			     i != attribute_list.end(); i++) {

				Glib::ustring attribute_name = (*i)->get_name();
				Glib::ustring attribute_value = (*i)->get_value();
				if (attribute_value.empty()) {
					throw ParserStyleError(attribute_name + _(": no value"));
				}

				switch (attribute_name[0]) {
				case 'i':	// id
					verify_attribute("id", attribute_name);
					if (attribute_value.length() > 7) {
						switch (attribute_value[0]) {
						case 'q':	// question_new
							verify_value("question_new", attribute_value);
							id = Audio::question_new;
							break;
						case 'a':	// answer_select + answer_wrong + answer_correct
							switch (attribute_value[7]) {
							case 's':	// answer_select + action_select
								if (attribute_value[1] == 'n') {	// answer_select
									verify_value("answer_select",
									             attribute_value);
									id = Audio::answer_select;
								} else {	// action_select
									verify_value("action_select",
									             attribute_value);
									id = Audio::action_select;
								}
								break;
							case 'w':	// answer_wrong
								verify_value("answer_wrong", attribute_value);
								id = Audio::answer_wrong;
								break;
							default:	// answer_correct
								verify_value("answer_correct", attribute_value);
								id = Audio::answer_correct;
								break;
							}
							break;
						}
					} else {
						throw ParserStyleError(
							  std::string(_("unknown attribute"))
							+ ": " + name);
					}
					break;
				case 'f':	// file
					verify_attribute("file", attribute_name);
					file = attribute_value;
					break;
				default:
					throw ParserStyleError(
						  std::string(_("unknown attribute"))
						+ ": " + attribute_name);

				}
			}

#ifndef HAVE_SOUND
			try {
#endif
				audio->set_sound(id, configuration->find_file(file));
#ifndef HAVE_SOUND
			} catch (const FileNotFound& e) {
				std::cerr << e.what()
					<< ' ' << _("(ignored)") << '\n';
			}
#endif
			break;
		}
		} // end switch

		xmlpp::Node::NodeList node_list = element->get_children();
		for (xmlpp::Node::NodeList::iterator i = node_list.begin();
		     i != node_list.end(); i++) {
			recurse_node_audio(*i, audio, configuration);
		}
	}
}



void
recurse_node_keys(const xmlpp::Node* node, Keys& keys,
	Configuration* configuration)
{
	if (const xmlpp::Element* element
	           	= dynamic_cast<const xmlpp::Element*>(node)) {
		Glib::ustring name = element->get_name();

		switch (name[0]) {
		case 'c': {	// command
			verify_tag("command", name);
			command_t id = command_undefined;
			guint keyval = 0;
			int number = 0;
			int pos = 0;

			const xmlpp::Element::AttributeList& attribute_list
				= element->get_attributes();

			for (xmlpp::Element::AttributeList::const_iterator i
			     	= attribute_list.begin();
			     i != attribute_list.end(); i++) {

				Glib::ustring attribute_name = (*i)->get_name();
				Glib::ustring attribute_value = (*i)->get_value();
				if (attribute_value.empty()) {
					throw ParserStyleError(attribute_name + _(": no value"));
				}

				switch (attribute_name[0]) {
				case 'i':	// id
					verify_attribute("id", attribute_name);

					switch (attribute_value[0]) {
					case 'n':	// next_question + next_level + next_status
						if (attribute_value.length() > 5) {
							switch (attribute_value[5]) {
							case 'q':	// next_question
								verify_value("next_question", attribute_value);
								id = command_next_question;
								break;
							case 'l':	// next_level
								verify_value("next_level", attribute_value);
								id = command_next_level;
								break;
							case 's':	// next_status
								verify_value("next_status", attribute_value);
								id = command_next_status_begin;
								break;
							default:
								id = command_undefined;
							}
						} else {
							id = command_undefined;
						}
						break;
					case 'p':	// prev_question + prev_level
						if (attribute_value.length() > 5) {
							switch (attribute_value[5]) {
							case 'q':	// prev_question
								verify_value("prev_question", attribute_value);
								id = command_prev_question;
								break;
							case 'l':	// prev_level
								verify_value("prev_level", attribute_value);
								id = command_prev_level;
								break;
							case 's':	// prev_status
								verify_value("prev_status", attribute_value);
								id = command_prev_status_begin;
								break;
							default:
								id = command_undefined;
							}
						} else {
							id = command_undefined;
						}
						break;
					case 't':	// toggle_fullscreen
						verify_value("toggle_fullscreen", attribute_value);
						id = command_toggle_fullscreen;
						break;
					case 'q':	// quit
						verify_value("quit", attribute_value);
						id = command_quit;
						break;
					case 'c':	// continue + continue_benevolent
						if (attribute_value.length() > 8) {
							verify_value("continue_benevolent", attribute_value);
							id = command_continue_benevolent;
						} else {
							verify_value("continue", attribute_value);
							id = command_continue;
						}
						break;
					case 's':	// select_answer + select_action + set_status
						if (attribute_value.length() > 8) {
							switch (attribute_value[8]) {
							case 'n':	// select_answer
								verify_value("select_answer", attribute_value);
								id = command_select_answer_begin;
								break;
							case 'c':	// select_action
								verify_value("select_action", attribute_value);
								id = command_select_action_begin;
								break;
							case 'u':	// set_status;
								verify_value("set_status", attribute_value);
								id = command_set_status_begin;
								break;
							default:
								id = command_undefined;
							}
						} else {
							id = command_undefined;
						}
						break;
					case 'd':	// deselect_answer
						verify_value("deselect_answer", attribute_value);
						id = command_deselect_answer_begin;
						break;
					default:
						id = command_undefined;
					}

					if (id == command_undefined) {
						throw ParserStyleError(
							  std::string(_("unknown command"))
							+ ": " + attribute_value);
					}

					break;
				case 'n':	// number
					verify_attribute("number", attribute_name);
					if (   id == command_select_answer_begin
					    || id == command_deselect_answer_begin
					    || id == command_select_action_begin
					    || id == command_next_status_begin
					    || id == command_prev_status_begin
					    || id == command_set_status_begin) {
						std::istringstream istr(attribute_value);
						istr >> number;

						if (number > 0) {
							number--;
						} else {
							throw ParserStyleError(
								  std::string(_("invalid number"))
								+ ": " + attribute_value);
						}
					} else {
						std::cerr << _("number ignored") << '\n';
					}

					break;
				case 'p':	// pos
					verify_attribute("pos", attribute_name);
					if (id == command_set_status_begin) {
						std::istringstream istr(attribute_value);
						istr >> pos;

						if (pos > 0) {
							pos--;
						} else {
							throw ParserStyleError(
								  std::string(_("invalid pos"))
								+ ": " + attribute_value);
						}
					} else {
						std::cerr << _("pos ignored") << '\n';
					}

					break;
				case 'k':	// key
					verify_attribute("key", attribute_name);

					if (attribute_value.length() == 1) {
						keyval = attribute_value[0];
					} else {
						keyval = get_key(attribute_value);
					}

					break;
				default:
					throw ParserStyleError(
						  std::string(_("unknown attribute"))
						+ ": " + attribute_name);

				}
			}

			id = command_t(id + number + pos);
			keys.set_key_command(keyval, id);
			break;
		} // end case 'c'
		case 'a': {	// actionkey
			verify_tag("actionkey", name);

			Glib::ustring plugin;
			Glib::ustring plugin_arg;
			Glib::ustring id;
			guint keyval = 0;

			const xmlpp::Element::AttributeList& attribute_list
				= element->get_attributes();

			for (xmlpp::Element::AttributeList::const_iterator i
			     	= attribute_list.begin();
			     i != attribute_list.end(); i++) {

				Glib::ustring attribute_name = (*i)->get_name();
				Glib::ustring attribute_value = (*i)->get_value();
				if (attribute_value.empty()) {
					throw ParserStyleError(attribute_name + _(": no value"));
				}

				switch (attribute_name[0]) {
				case 'i':	// id
					verify_attribute("id", attribute_name);
					id = attribute_value;

					break;
				case 'k':	// key
					verify_attribute("key", attribute_name);

					if (attribute_value.length() == 1) {
						keyval = attribute_value[0];
					} else {
						keyval = get_key(attribute_value);
					}

					break;
				case 'p':	// plugin + pluginarg
					if (attribute_name.length() > 6) {
						verify_attribute("pluginarg", attribute_name);
						plugin_arg = attribute_value;
					} else {
						verify_attribute("plugin", attribute_name);
						plugin = attribute_value;
					}

					break;
				default:
					throw ParserStyleError(
						  std::string(_("unknown attribute"))
						+ ": " + attribute_name);
				}
			}

			command_t command = command_t(
				  int(command_select_action_autonomous_begin)
				+ configuration->add_action_autonomous(plugin, plugin_arg));
			keys.set_key_command(keyval, command);
			break;
		}
		default:
			throw ParserStyleError(
				  std::string(_("unknown tag"))
				+ ": " + name);
		} // end switch name[0]

		xmlpp::Node::NodeList node_list = element->get_children();
		for (xmlpp::Node::NodeList::iterator i = node_list.begin();
		     i != node_list.end(); i++) {
			recurse_node_keys(*i, keys, configuration);
		}
	}
}



void
recurse_node(const xmlpp::Node* node, Configuration* configuration)
{
	if (const xmlpp::Element* element
	           	= dynamic_cast<const xmlpp::Element*>(node)) {
		Glib::ustring name = element->get_name();

		switch (name[0]) {
		case 'a': {	// audio
			verify_tag("audio", name);

			xmlpp::Node::NodeList node_list = element->get_children();
			for (xmlpp::Node::NodeList::iterator i = node_list.begin();
			     i != node_list.end(); i++) {
				recurse_node_audio(*i, configuration->audio(), configuration);
			}

			break;
		}
		case 'k': {	// keys
			verify_tag("keys", name);

			xmlpp::Node::NodeList node_list = element->get_children();
			for (xmlpp::Node::NodeList::iterator i = node_list.begin();
			     i != node_list.end(); i++) {
				recurse_node_keys(*i, configuration->keys(), configuration);
			}

			break;
		}
		case 'g': {	// group
			verify_tag("group", name);

			const xmlpp::Element::AttributeList& attribute_list
				= element->get_attributes();
			group_t group = group_trivia;

			for (xmlpp::Element::AttributeList::const_iterator i
			     	= attribute_list.begin();
			     i != attribute_list.end(); i++) {

				Glib::ustring attribute_name = (*i)->get_name();
				Glib::ustring attribute_value = (*i)->get_value();
				if (attribute_value.empty()) {
					throw ParserStyleError(attribute_name + _(": no value"));
				}

				switch (attribute_name[0]) {
				case 'i':	// id
					verify_attribute("id", attribute_name);

					switch (attribute_value[0]) {
					case 'i':	// intro
						verify_value("intro", attribute_value);
						group = group_intro;
						break;
					case 't':	// trivia
						verify_value("trivia", attribute_value);
						group = group_trivia;
						break;
					case 'w':	// winoutro
						verify_value("winoutro", attribute_value);
						group = group_outro_win;
						break;
					case 'l':	// loseoutro
						verify_value("loseoutro", attribute_value);
						group = group_outro_lose;
						break;
					default:
						throw ParserStyleError(
							  std::string(_("unknown group id")) + ": "
							+ attribute_value);
					}
				}
			}

			xmlpp::Node::NodeList node_list = element->get_children();
			for (xmlpp::Node::NodeList::iterator i = node_list.begin();
			     i != node_list.end(); i++) {
				recurse_node_display(*i, configuration->item_top(), group,
					configuration);
			}

			break;
		}
		default: {
			xmlpp::Node::NodeList node_list = element->get_children();
			for (xmlpp::Node::NodeList::iterator i = node_list.begin();
			     i != node_list.end(); i++) {
				recurse_node(*i, configuration);
			}
		}
		}
	} // end if element
}



bool
parse_style(Configuration& configuration)
{
	try {
		xmlpp::DomParser parser;

		parser.set_substitute_entities();
		parser.parse_file(configuration.style_file());

		if (parser) {
			const xmlpp::Node* node
				= parser.get_document()->get_root_node();

			recurse_node(node, &configuration);

			return true;
		} else {
			throw ParserStyleError(_("Could not parse file"));
		}
	} catch (const FileNotFound&) {
		throw;
	} catch (const ParserError&) {
		throw;
	} catch (const std::bad_alloc&) {
		throw;
	} catch (const std::exception& e) {
		throw ParserStyleError(e.what());
	} catch (...) {
		throw;
	}

}





} // end namespace quimeleon
